#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#P:/64_PGM/64_PYT/m27__DescPropSlot.py
"""
Příkazy a záznamy odpovědí zadávané ve výpisech kapitoly:
27  Deskriptory, vlastnosti a sloty.

Co se v kapitole naučíte
Tato kapitola představí deskriptory a vlastnosti. Probere možnosti
jejich zadání a použití a upozorní na některá jejich specifika.
Poté naznačí, jak se rozhodovat, kterému z probraných způsobů přístupu
k atributům dát v dané situaci přednost.
Na závěr ještě představí sloty, jako další způsob definice atributů.

Společné konvence dodržované v doprovodných programech jsou popsány
v modulu m01__Startujeme.
"""
from dbg import DBG
if DBG>0: print(f'===== Modul {__name__} ===== START')
############################################################################



#Výpis 27.1: Definice třídy descriptor
############################################################################
# >>> \
class descriptor:
    """Datový deskriptor průběžně zaznamenávající jednotlivé přístupy."""

    def __init__(self, initval=None):
        print(f'Vytvořen deskriptor s výchozí hodnotou «{initval}»')
        self.pool = {None: initval}

    def __set_name__(self, owner, name):
        self.name = name
        print('Přiřazuje se název:', name, '-  Vlastník:', owner)

    def __get__(self, instance, objtype):
        val = self                if instance is None else \
              self.pool[instance] if instance in self.pool.keys() else \
              self.pool[None]
        print(f"Čtení: atribut {self.name} objektu {instance} "
              f"typu {objtype} = {val}")
        return val

    def __set__(self, instance, val):
        print(f"Zápis atributu {self.name} objektu {instance} = {val}")
        self.pool[instance] = val

    def __delete__(self, instance):
        print(f"Odstranění atributu {self.name} objektu {instance}")
        self.pool.pop(instance)

# >>>



#Výpis 27.2: Hrátky s deskriptorem
############################################################################
# >>> \
class C27a:
    a1 = descriptor('První')
    a2 = descriptor('Druhý')
    def __init__(self, id):   self.id = id
    def __repr__(self):   return 'C27a_' + str(self.id)
    print('Třída C27a definována')

# Vytvořen deskriptor s výchozí hodnotou «První»
# Vytvořen deskriptor s výchozí hodnotou «Druhý»
# Třída C27a definována
# Přiřazuje se název: a1 -  Vlastník: <class '__main__.C27a'>
# Přiřazuje se název: a2 -  Vlastník: <class '__main__.C27a'>
c1 = C27a(1);   c2 = C27a(2);   C27a.a1,   c1.a1,   c2.a1
# Čtení: atribut a1 objektu None typu <class '__main__.C27a'> = <__main__.descriptor object at 0x0000028E1AE45880>
# Čtení: atribut a1 objektu C27a_1 typu <class '__main__.C27a'> = První
# Čtení: atribut a1 objektu C27a_2 typu <class '__main__.C27a'> = První
# (<__main__.descriptor object at 0x0000028E1AE45880>, 'První', 'První')
c1.a1='C1';   c2.a1='C2';   C27a.a1, c1.a1, c2.a1, c1.a2, c2.a2
# Zápis atributu a1 objektu C27a_1 = C1
# Zápis atributu a1 objektu C27a_2 = C2
# Čtení: atribut a1 objektu None typu <class '__main__.C27a'> = <__main__.descriptor object at 0x0000028E1AE45880>
# Čtení: atribut a1 objektu C27a_1 typu <class '__main__.C27a'> = C1
# Čtení: atribut a1 objektu C27a_2 typu <class '__main__.C27a'> = C2
# Čtení: atribut a2 objektu C27a_1 typu <class '__main__.C27a'> = Druhý
# Čtení: atribut a2 objektu C27a_2 typu <class '__main__.C27a'> = Druhý
# (<__main__.descriptor object at 0x0000028E1AE45880>, 'C1', 'C2', 'Druhý', 'Druhý')
C27a.a1='Třídní';   C27a.a1, c1.a1, c2.a1, c1.a2, c2.a2
# Čtení: atribut a2 objektu C27a_1 typu <class '__main__.C27a'> = Druhý
# Čtení: atribut a2 objektu C27a_2 typu <class '__main__.C27a'> = Druhý
# ('Třídní', 'Třídní', 'Třídní', 'Druhý', 'Druhý')
# >>>



#Výpis 27.3: Deskriptor vytvářený vnitřní třídou
############################################################################
# >>> \
class C27b:
    def __init__(self, val):
        self._atr = val

    class descr:
        def __get__(self, instance, owner):
            return f"Přepočtená hodnota «{instance._atr}»"
        def __set__(self, instance, value):
            raise AttributeError('Atribut atr nelze měnit')
        def __delete__(self, instance, value):
            raise AttributeError('Atribut atr nelze odstranit')

    atr = descr()

cc1=C27b('První');   cc2=C27b('Druhý')
cc1.atr,   cc2.atr
# ('Přepočtená hodnota «První»', 'Přepočtená hodnota «Druhý»')
cc1.atr = 0
# Traceback (most recent call last):
#   File "<pyshell#10>", line 1, in <module>
#     cc1.atr = 0
#   File "<pyshell#7>", line 10, in __set__
#     raise AttributeError('Atribut atr nelze měnit')
# AttributeError: Atribut atr nelze měnit
# >>>



#Výpis 27.4: Deskriptor pro odložené vyhodnocení
############################################################################
# >>> \
class reify(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val

# >>> \
class C27c:
    @reify
    def attr(self):
        print('Počítám hodnotu atributu...')
        return 'Spočteno'

cc = C27c();   cc.attr
# Počítám hodnotu atributu...
# 'Spočteno'
cc.attr, cc.attr
# ('Spočteno', 'Spočteno')
# >>>



#Výpis 27.5 Vlastnost jako instance třídy property
############################################################################
# >>> \
class C27d:
    def __init__(self, id, val):   self._id = id;   self._attr = val
    def __repr__(self):      return f"C27d({self._id})"
    def getAttr (self):      print(f"Čtení {self}:"); return self._attr
    def setAttr (self, val): print(f"Zápis {self}:"); self._attr = val
    # def delAttr (self):      print(f"Mazat {self}:"); del self._attr
    # attr = property(getAttr, setAttr, delAttr, 'Demonstrační atribut')
    attr = property(getAttr, setAttr, doc='Demonstrační atribut')

c1 = C27d(1, 'Raz');   c2 = C27d(2, 'Dva');   c1, c2
# (C27d(1), C27d(2))
c1.attr='One';   c2.attr='Two';   c1.attr, c2.attr
# Zápis C27d(1):
# Zápis C27d(2):
# Čtení C27d(1):
# Čtení C27d(2):
# ('One', 'Two')
help(C27d.attr)
# Help on property:

#     Demonstrační atribut

# >>>



#Výpis 27.6: Ekvivalent definice vestavěné třídy property v modulu m27b_Property
############################################################################

""" Je uloženo v souboru m27b_Property.py. """



#Výpis 27.7: Zadání vlastnosti pomocí dekorátoru
############################################################################
# >>> \
class C27e:
    def __init__(self, id, val):
        self._id = id;
        self._attr = val

    def __repr__(self):
        return f"C27d({self._id}-{self._attr})"

    @property
    def attr(self):
        """Dokumentační komentář vlastnosti C27e.attr."""
        print(f"Čtení {self}.attr (= {self._attr})");
        return self._attr

    @attr.setter
    def attr(self, val):
        print(f"Zápis {self}.attr ({self._attr}->{val})")
        self._attr = val

    @attr.deleter
    def attr(self):
        print(f"Mazat {self}.attr (= {self._attr})")
        del self._attr

c1 = C27e(1, 'Raz');   c2 = C27e(2, 'Dva');   c1, c2
# ((C27d(1-Raz), C27d(2-Dva))
c1.attr='One';   c2.attr='Two';   c1.attr, c2.attr
# Zápis C27d(1-Raz).attr (Raz->One)
# Zápis C27d(2-Dva).attr (Dva->Two)
# Čtení C27d(1-One).attr (= One)
# Čtení C27d(2-Two).attr (= Two)
# ('One', 'Two')
help(C27e.attr)
# Help on property:

    # Dokumentační komentář vlastnosti C27e.attr.

# >>>



#Výpis 27.8: Ukázka používání slotů
############################################################################
# >>> \
class Point:
    __slots__ = ("x", "y")

    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __repr__(self):
        return f"Point(x={self.x}, y={self.y})"

p = Point(1, 2);   p
# Point(x=1, y=2)
p.x = 3;   p
# Point(x=3, y=2)
p.z = 4
# Traceback (most recent call last):
#   File "<pyshell#33>", line 1, in <module>
#     p.z = 4
# AttributeError: 'Point' object has no attribute 'z'
# >>>



#Výpis 27.Y: Popisek
############################################################################
if DBG>0: print(f'===== Modul {__name__} ===== STOP')
